/*
    SDL - Simple DirectMedia Layer
    Copyright (C) 1997-2006 Sam Lantinga

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Sam Lantinga
    slouken@libsdl.org
*/
#include <sys/ioctl.h>
#include <fcntl.h>

#include "SDL_config.h"
#include "SDL_wiz_dev.h"

#include "simple_isa1200if.h"

#define BATT_DEV 	"/dev/misc/pollux_batt"
#define BATT_DEV_WITH_DEVFS "/dev/pollux_batt"
#define VIB_DEV "/dev/isa1200"
#define GPIO_DEV "/dev/GPIO"

struct SDL_WizDevif WizDevif = {-1, -1, -1};

static pattern_data_t vib_pattern={
	.act_number = 6,
	.vib_act_array = {{0,126},{100,-126},{120,0},{150,126},{250,-126},{265,0}}
};

/* Battery */
int SDL_WizDevIfInit(Uint32 init)
{
	//printf("SDL_WizDevIfInit(init=0x%x)\n", init);
#if 0 /* modified by ikari 2010.3.3 */
	if( ( (init & 0xff) & SDL_WIZ_BATTRY_INIT) == SDL_WIZ_BATTRY_INIT )
	{

		WizDevif.batt_fd = open(BATT_DEV, (O_RDWR | ( (init >>16) & 0xffff))  );

		if(WizDevif.batt_fd < 0){
			WizDevif.batt_fd = open(BATT_DEV_WITH_DEVFS, (O_RDWR | ( (init >>16) & 0xffff))  );

			if (WizDevif.batt_fd < 0){
			    printf("open battery driver failed\n");
			    return -1 ;
			}
		}               
	}
#endif

	if( ( (init & 0xff) & SDL_WIZ_BATTRY_INIT) == SDL_WIZ_BATTRY_INIT )
	{
		WizDevif.batt_fd = open(BATT_DEV, (O_RDONLY | ( (init >>16) & 0xffff)));

		if(WizDevif.batt_fd < 0){
			WizDevif.batt_fd = open(BATT_DEV_WITH_DEVFS, (O_RDONLY | ( (init >>16) & 0xffff)));

			if (WizDevif.batt_fd < 0){
				printf("open battery driver failed\n");
				return -1 ;
			}
		}
	}
	return 0;
}

int SDL_GetBattryCheck(Uint16 *battry_status)
{
    if( WizDevif.batt_fd < 0 ) return -1;
    
    //read(WizDevif.batt_fd, (char *)battry_status, sizeof(Uint16));
    read(WizDevif.batt_fd, battry_status, sizeof(Uint16));
    
    return 0;
}

/* modified by ikari 2010.3.3 */
int SDL_WizDevIfQuit(void)
{
	if(WizDevif.batt_fd>=0)
		close(WizDevif.batt_fd);
	return 0;
}

/* Vibrate */
int SDL_VibrateInit(void)
{
	if(WizDevif.vib_fd<0)
		WizDevif.vib_fd=open(VIB_DEV, O_RDWR|O_NDELAY);

	if(WizDevif.vib_fd<0){
		printf("open vib driver failed\n");
		return -1;
	}

	return 0;
}

int SDL_VibrateQuit(void)
{
	if(WizDevif.vib_fd>=0)
		close(WizDevif.vib_fd);
	return 0;
}

int SDL_VibratePlay(int _value)
{
	if(WizDevif.vib_fd<0) return -1;
	ioctl(WizDevif.vib_fd,IOCTL_SET_VIB_LEVEL,&_value);
	ioctl(WizDevif.vib_fd,IOCTL_PLAY_PATTERN,&vib_pattern);
	return 0;
}

/* GPIO IOCTRL */
int SDL_OpenGPIO(void)
{
	if(WizDevif.gpio_fd<0) 
		WizDevif.gpio_fd=open(GPIO_DEV, O_RDWR | O_NDELAY);

	if (WizDevif.gpio_fd < 0){
		printf("open gpio driver failed\n");
		return -1 ;
	}

	return 0;
}

int SDL_CloseGPIO(void)
{
	if(WizDevif.gpio_fd>=0)
		close(WizDevif.gpio_fd);
	return 0;
}

int SDL_SDStatus(void)
{
	int status=-1;
	if(WizDevif.gpio_fd>=0)
		ioctl(WizDevif.gpio_fd,IOCTL_GET_INSERT_SDMMC,&status);
	return status;
}

int SDL_HoldStatus(void)
{
	int status=-1;
	if(WizDevif.gpio_fd>=0)
		ioctl(WizDevif.gpio_fd,IOCTL_GET_HOLD_STATUS,&status);
	return status;
}

int SDL_USBWLANStatus(void)
{
	int status=-1;
	if(WizDevif.gpio_fd>=0)
		ioctl(WizDevif.gpio_fd,IOCTL_GET_USB_WIRELESS_LAN,&status);
	return status;
}

int SDL_SYS_JoystickUsbConCheck(void)
{
	int status=-1;
	if(WizDevif.gpio_fd>=0)
		ioctl(WizDevif.gpio_fd,IOCTL_GET_USB_CHECK,&status);
	return status;
}

int SDL_SYS_JoystickUsbDisconnect(void)
{
	if(WizDevif.gpio_fd>=0)
		ioctl(WizDevif.gpio_fd,IOCTL_SET_SOFT_DISCONNECT,NULL);
	return 0;
}

int SDL_USBSETStatus(void)
{
	unsigned int status=0;
	if(WizDevif.gpio_fd<0) return -1;
		ioctl(WizDevif.gpio_fd,IOCTL_SET_USB_MEMORY_STICK_MOUNT,&status);
	return 0;
}

int SDL_USBGETStatus(void)
{
	int status=-1;
	if(WizDevif.gpio_fd>=0)
		ioctl(WizDevif.gpio_fd,IOCTL_GET_USB_MEMORY_STICK_MOUNT,(unsigned int*)&status);
	return status;
}

int SDL_USBENABLEStatus(void)
{
	int status=-1;
	if(WizDevif.gpio_fd>=0)
		ioctl(WizDevif.gpio_fd,IOCTL_GET_USB_MEMORY_STICK_ENABLE,(unsigned int*)&status);
	return status;
}

