/*
    SDL - Simple DirectMedia Layer
    Copyright (C) 1997-2006 Sam Lantinga

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Sam Lantinga
    slouken@libsdl.org
*/
#include "SDL_config.h"

/* SDL Caanoo video driver implementation.
   Copied from SDL-1.2.13/src/video/fbcon/SDL_fbvideo.c and modified it.
   2010.1.23 ikari
*/

#include "SDL_video.h"
#include "SDL_mouse.h"
#include "../SDL_sysvideo.h"
#include "../SDL_pixels_c.h"
#include "../../events/SDL_events_c.h"
#include "SDL_caanoovideo.h"
#include "SDL_caanoomouse_c.h"
#include "SDL_caanooevents_c.h"
#include "SDL_caanooyuv.h"

#include <sys/ioctl.h>
#include <sys/mman.h>
#include <unistd.h>
#include <fcntl.h>

#include "polluxregs.h"

#define CAANOO_FB_LENGTH 0x100000

#define YUV_MEM_START 0x5800000
#define YUV_MEM_LENGTH 0x1800000

#define CAANOO_FlipFB(back) (MLCADDRESS1=this->fb1_realaddr[back])
#define CAANOO_FlipFB_TV(back) (MLCADDRESS1_TV=this->fb1_realaddr[back])


// ghcstop,  SDL_sysideo.h:  #define _THIS        SDL_VideoDevice *_this

/* ghcstop_041124 add
 *
 * keyboard, mouse input ִ define, 1̸ enable, 0̸ disable
 *
 * mplayer 20041123  ̺κ 0 ϰ  
 *  Ĺ ̺κ 1 ϰ SDL_fbevents.c #define FB_VT_OPEN 1 // ghcstop add
 * - testbitmap.c ߵȴ...--;, 20041124) 
 */
#define FB_INPUT_DEVICE_ENABLE 1 

/* HYUN_DEBUG */
//extern int SDL_videofd; // Dark add 20050510

/* Initialization/Query functions */
static int CAANOO_VideoInit(_THIS, SDL_PixelFormat *vformat);
static SDL_Rect **CAANOO_ListModes(_THIS, SDL_PixelFormat *format, Uint32 flags);
static SDL_Surface *CAANOO_SetVideoMode(_THIS, SDL_Surface *current, int width, int height, int bpp, Uint32 flags);
static int CAANOO_SetColors(_THIS, int firstcolor, int ncolors, SDL_Color *colors);
static void _CAANOO_VideoQuit(_THIS);
static void CAANOO_VideoQuit(_THIS);

/* Hardware surface functions */
static int CAANOO_InitHWSurfaces(_THIS, SDL_Surface *screen, unsigned char *base, int size);
static void CAANOO_FreeHWSurfaces(_THIS);
static int CAANOO_AllocHWSurface(_THIS, SDL_Surface *surface);
static int CAANOO_LockHWSurface(_THIS, SDL_Surface *surface);
static void CAANOO_UnlockHWSurface(_THIS, SDL_Surface *surface);
static void CAANOO_FreeHWSurface(_THIS, SDL_Surface *surface);
static int CAANOO_FlipHWSurface(_THIS, SDL_Surface *surface);
static int CAANOO_FlipHWSurface_TV(_THIS, SDL_Surface *surface);
static int CAANOO_FlipHWSurface_Both(_THIS, SDL_Surface *surface);


static int CAANOO_Available(void)
{	
	return 1;
}

static void CAANOO_DeleteDevice(SDL_VideoDevice *device)
{
	SDL_free(device->hidden);
	SDL_free(device);
}

static SDL_VideoDevice *CAANOO_CreateDevice(int devindex)
{
	SDL_VideoDevice *this;

	/* Initialize all variables that we clean on shutdown */
	this = (SDL_VideoDevice *)SDL_malloc(sizeof(SDL_VideoDevice));
	if ( this ) {
		SDL_memset(this, 0, (sizeof *this));
		this->hidden = (struct SDL_PrivateVideoData *)
				SDL_malloc((sizeof *this->hidden));
	}
	if ( (this == NULL) || (this->hidden == NULL) ) {
		SDL_OutOfMemory();
		if ( this ) {
			SDL_free(this);
		}
		return(0);
	}
	SDL_memset(this->hidden, 0, (sizeof *this->hidden));

	mem_fd = -1;
	mouse_fd = -1;
	keyboard_fd = -1;
	bkregs32[0]=NULL;
	bkregs32[1]=NULL;

	/* Set the function pointers */
	this->VideoInit = CAANOO_VideoInit;
	this->ListModes = CAANOO_ListModes;
	this->SetVideoMode = CAANOO_SetVideoMode;
	this->SetColors = CAANOO_SetColors;
	this->UpdateRects = NULL;
	this->CreateYUVOverlay = CAANOO_CreateYUVOverlay;
	this->VideoQuit = CAANOO_VideoQuit;
	this->AllocHWSurface = CAANOO_AllocHWSurface;
	this->CheckHWBlit = NULL;
	this->FillHWRect = NULL;
	this->SetHWColorKey = NULL;
	this->SetHWAlpha = NULL;
	this->LockHWSurface = CAANOO_LockHWSurface;
	this->UnlockHWSurface = CAANOO_UnlockHWSurface;
	this->FlipHWSurface = CAANOO_FlipHWSurface;
	this->FreeHWSurface = CAANOO_FreeHWSurface;
	this->SetCaption = NULL;
	this->SetIcon = NULL;
	this->IconifyWindow = NULL;
	this->GrabInput = NULL;
	this->GetWMInfo = NULL;
	this->InitOSKeymap = CAANOO_InitOSKeymap;
	this->PumpEvents = CAANOO_PumpEvents;

	this->free = CAANOO_DeleteDevice;

	return this;
}

VideoBootStrap CAANOO_bootstrap = {
	"caanoo", "GPH Caanoo SDL video driver",
	CAANOO_Available, CAANOO_CreateDevice
};

static int CAANOO_VideoInit(_THIS, SDL_PixelFormat *vformat)
{
	int i;
	struct fb_fix_screeninfo info;

	/* Open screen framebuffer */
	fb_fd=open("/dev/fb0",O_RDWR);
	if(fb_fd<0) 
	{
		SDL_SetError("Couldn't open /dev/fb0");
		CAANOO_VideoQuit(this);
		return(-1);
	}

	if(ioctl(fb_fd,FBIOGET_FSCREENINFO,&info)<0) 
	{
		SDL_SetError("Couldn't read framebuffer information");
		CAANOO_VideoQuit(this);
		return(-1);
	}

	/* Open /dev/mem */
	mem_fd=open("/dev/mem",O_RDWR);
	if(mem_fd<0)
	{
		SDL_SetError("Couldn't open GP2X mem device");
		CAANOO_VideoQuit(this);
		return(-1);
	}

#if !SDL_THREADS_DISABLED
	/* Create the hardware surface lock mutex */
	hw_lock = SDL_CreateMutex();
	if ( hw_lock == NULL ) {
		SDL_SetError("Unable to create lock mutex");
		CAANOO_VideoQuit(this);
		return(-1);
	}
#endif

	/* mmap CAANOO hardware register */
	memregs32=(volatile unsigned int*)mmap(0,0x20000,PROT_READ|PROT_WRITE,MAP_SHARED,mem_fd,0xC0000000);
	if(memregs32==(volatile unsigned int*)0xFFFFFFFF)
	{
		SDL_SetError("Couldn't mmap Caanoo hardware register");
		CAANOO_VideoQuit(this);
		return(-1);
	}
	memregs16=(volatile unsigned short*)memregs32;

	yuv_mem_start=(unsigned char*)mmap(0,YUV_MEM_LENGTH,PROT_READ|PROT_WRITE,MAP_SHARED,mem_fd,YUV_MEM_START);
	if(yuv_mem_start==(unsigned char*)0xFFFFFFFF)
	{
		SDL_SetError("Couldn't mmap 40MiB YUV area.");
		CAANOO_VideoQuit(this);
		return(-1);
	}

	fb1[0]=(unsigned char*)mmap(0,CAANOO_FB_LENGTH,PROT_READ|PROT_WRITE,MAP_SHARED,mem_fd,info.smem_start);
	if(fb1[0]==(unsigned char*)0xFFFFFFFF)
	{
		SDL_SetError("Couldn't mmap framebuffer memory.");
		CAANOO_VideoQuit(this);
		return(-1);
	}
	this->fb1_realaddr[0]=info.smem_start;

	/* Determine TV-out mode */
	this->tvout_mode=0;
	if(DPCCTRL0 & 0x8000) 
		this->tvout_mode |= MLC_PRI_ENABLED;
	if(DPCCTRL0_TV & 0x8000) 
		this->tvout_mode |= MLC_SEC_ENABLED;

	/* Backup previous register values */
	bkregs32[0]=(unsigned int*)malloc(sizeof(unsigned int)*22);
	if(!bkregs32[0])
	{
		SDL_SetError("Not enough memory.");
		CAANOO_VideoQuit(this);
		return(-1);
	}

	bkregs32[0][0]=MLCADDRESS1;
	bkregs32[0][1]=MLCCONTROL1;
	bkregs32[0][2]=MLCLEFTRIGHT1;
	bkregs32[0][3]=MLCTOPBOTTOM1;
	bkregs32[0][4]=MLCLEFTRIGHT1_0;
	bkregs32[0][5]=MLCTOPBOTTOM1_0;
	bkregs32[0][6]=MLCLEFTRIGHT1_1;
	bkregs32[0][7]=MLCTOPBOTTOM1_1;
	bkregs32[0][8]=MLCBGCOLOR;
	bkregs32[0][9]=MLCHSTRIDE1;
	bkregs32[0][10]=MLCVSTRIDE1;
	bkregs32[0][11]=MLCADDRESS2;
	bkregs32[0][12]=MLCADDRESSCB;
	bkregs32[0][13]=MLCADDRESSCR;
	bkregs32[0][14]=MLCCONTROL2;
	bkregs32[0][15]=MLCLEFTRIGHT2;
	bkregs32[0][16]=MLCTOPBOTTOM2;
	bkregs32[0][17]=MLCVSTRIDE2;
	bkregs32[0][18]=MLCVSTRIDECB;
	bkregs32[0][19]=MLCVSTRIDECR;
	bkregs32[0][20]=MLCHSCALE;
	bkregs32[0][21]=MLCVSCALE;

	bkregs32[1]=(unsigned int*)malloc(sizeof(unsigned int)*22);
	if(!bkregs32[1])
	{
		SDL_SetError("Not enough memory.");
		CAANOO_VideoQuit(this);
		return(-1);
	}

	bkregs32[1][0]=MLCADDRESS1_TV;
	bkregs32[1][1]=MLCCONTROL1_TV;
	bkregs32[1][2]=MLCLEFTRIGHT1_TV;
	bkregs32[1][3]=MLCTOPBOTTOM1_TV;
	bkregs32[1][4]=MLCLEFTRIGHT1_0_TV;
	bkregs32[1][5]=MLCTOPBOTTOM1_0_TV;
	bkregs32[1][6]=MLCLEFTRIGHT1_1_TV;
	bkregs32[1][7]=MLCTOPBOTTOM1_1_TV;
	bkregs32[1][8]=MLCBGCOLOR_TV;
	bkregs32[1][9]=MLCHSTRIDE1_TV;
	bkregs32[1][10]=MLCVSTRIDE1_TV;
	bkregs32[1][11]=MLCADDRESS2_TV;
	bkregs32[1][12]=MLCADDRESSCB_TV;
	bkregs32[1][13]=MLCADDRESSCR_TV;
	bkregs32[1][14]=MLCCONTROL2_TV;
	bkregs32[1][15]=MLCLEFTRIGHT2_TV;
	bkregs32[1][16]=MLCTOPBOTTOM2_TV;
	bkregs32[1][17]=MLCVSTRIDE2_TV;
	bkregs32[1][18]=MLCVSTRIDECB_TV;
	bkregs32[1][19]=MLCVSTRIDECR_TV;
	bkregs32[1][20]=MLCHSCALE_TV;
	bkregs32[1][21]=MLCVSCALE_TV;

	/* Fill initial values */
	fb1[1]=NULL;
	fb1[2]=NULL;
	fb1[3]=NULL;

	vformat->BitsPerPixel=16;
	vformat->BytesPerPixel=2;
	vformat->Rmask=0xf800;
	vformat->Gmask=0x07e0;
	vformat->Bmask=0x001f;
	vformat->Amask=0;

	/* Fill in our hardware acceleration capabilities */
	this->info.current_w = 320;
	this->info.current_h = 240;
	this->info.wm_available = 0;
	this->info.hw_available = 0;

	SDL_modelist[0]=SDL_malloc(sizeof(SDL_Rect));
	SDL_modelist[0]->x=0; SDL_modelist[0]->y=0;
	SDL_modelist[0]->w=320,SDL_modelist[0]->h=240;
	SDL_modelist[1]=NULL;

	/* Enable mouse and keyboard support */
	if ( CAANOO_OpenKeyboard(this) < 0 ) {
		CAANOO_VideoQuit(this);
		return(-1);
	}
	#if 0 // ghcstop delete, no keyboard, no mouse in s3c2410
	// ghcstop_041123: SDL_fbevents.c,  콺 츮 ߿ ߰ڱ...
	if ( FB_OpenMouse(this) < 0 ) {
		const char *sdl_nomouse;

		sdl_nomouse = getenv("SDL_NOMOUSE");
		if ( ! sdl_nomouse ) {
			SDL_SetError("Unable to open mouse");
			FB_VideoQuit(this);
			return(-1);
		}
	}
	#else
	CAANOO_OpenMouse(this);
       #endif

	/* We're done! */
	return(0);
}

static SDL_Rect **CAANOO_ListModes(_THIS, SDL_PixelFormat *format, Uint32 flags)
{
	return(SDL_modelist);
}

/* Various screen update functions available */
static void CAANOO_DirectUpdate(_THIS, int numrects, SDL_Rect *rects);

static SDL_Surface *CAANOO_SetVideoMode(_THIS, SDL_Surface *current,
				int width, int height, int bpp, Uint32 flags)
{
	Uint32 Rmask,Gmask,Bmask,Amask=0;
	Uint32 mode,bytesperpixel,hwsurface_length;
	char* env=NULL;

#if FB_INPUT_DEVICE_ENABLE // ghcstop_041123 alive
	/* Set the terminal into graphics mode */
	if ( CAANOO_EnterGraphicsMode(this) < 0 ) {
		return(NULL);
	}
#endif

	#if 0 // ghcstop_041123 delete
	/* Restore the original palette */
	FB_RestorePalette(this);
	#endif

#ifdef CAANOO_MPLAYER
	env=SDL_getenv("CAANOO_ARGB");
#endif
	switch(bpp)
	{
		case 8: /* 256 color */
			Rmask=255;
			Gmask=255;
			Bmask=255;
			mode=0x443AD020;
			fb1_length=0x12C00; /* 320*240 */
			break;
		case 15: /* RGB555 - Unusual */
			Rmask=0x7c00;
			Gmask=0x03e0;
			Bmask=0x001f;
			mode=0x4342D020;
			fb1_length=0x25800; /* 320*240*2 */
			break;
		case 16: /* RGB565 or ARGB4444 */
#ifdef CAANOO_MPLAYER
			if(env && (strcmp(env,"1")==0))
			{
				Rmask=0x0F00;
				Gmask=0x00F0;
				Bmask=0x000F;
				Amask=0xF000;
				mode=0x2211D024;
			}
			else
#endif
			{
				Rmask=0xf800;
				Gmask=0x07e0;
				Bmask=0x001f;
				mode=0x4432D020;
			}

			fb1_length=0x25800; /* 320*240*2 */
			break;
		case 24: /* RGB888 - It's too slow */
			Rmask=0xff0000;
			Gmask=0x00ff00;
			Bmask=0x0000ff;
			mode=0x4653D020;
			fb1_length=0x38400; /* 320*240*3 */
			break;
		case 32: /* XRGB8888 or ARGB8888 */
#ifdef CAANOO_MPLAYER
			if(env && (strcmp(env,"1")==0))
			{
				Rmask=0x00ff0000;
				Gmask=0x0000ff00;
				Bmask=0x000000ff;
				Amask=0xff000000;
				mode=0x0653D024;
			}
			else
#endif
			{
				Rmask=0x00ff0000;
				Gmask=0x0000ff00;
				Bmask=0x000000ff;
				mode=0x4653D020;
			}
			fb1_length=0x4B000; /* 320*240*4 */
			break;
		default:
			SDL_SetError("Not supported bpp");
			CAANOO_VideoQuit(this);
			return(NULL);
	}

	if ( ! SDL_ReallocFormat(current, bpp,
	                                  Rmask, Gmask, Bmask, Amask) ) {
		SDL_SetError("Couldn't set SDL videosurface");
		CAANOO_VideoQuit(this);
		return(NULL);
	}
	bytesperpixel=current->format->BytesPerPixel;

	/* Set up the new mode framebuffer */
	current->flags = (SDL_FULLSCREEN|SDL_HWSURFACE) | (flags & SDL_DOUBLEBUF);
	current->pixels = (void*)fb1[0];
	current_fb=0;

	/* Set up double buffer */
	if((flags & SDL_DOUBLEBUF)==SDL_DOUBLEBUF)
	{
		memset((void*)fb1[0],0x00,fb1_length*3);
		this->fb1_realaddr[1]=this->fb1_realaddr[0]+fb1_length;
		fb1[1]=fb1[0]+fb1_length;
		this->fb1_realaddr[2]=this->fb1_realaddr[1]+fb1_length;
		fb1[2]=fb1[1]+fb1_length;

		if(bpp==32)
		{
			this->fb1_realaddr[3]=YUV_MEM_START;
			fb1[3]=yuv_mem_start;
			hw_surface=fb1[0]+(fb1_length*3);
			hwsurface_length=CAANOO_FB_LENGTH-(fb1_length*3);
		}
		else
		{
			this->fb1_realaddr[3]=this->fb1_realaddr[2]+fb1_length;
			fb1[3]=fb1[2]+fb1_length;
			hw_surface=fb1[0]+(fb1_length*4);
			hwsurface_length=CAANOO_FB_LENGTH-(fb1_length*4);
		}
		memset((void*)fb1[3],0x00,fb1_length);

		if((flags & SDL_HWSURFACE)==SDL_HWSURFACE)
			current->pixels = (void*)fb1[1];

		current->flags |= (flags & SDL_VSYNCOFF);
	}
	else
	{
		memset((void*)fb1[0],0x00,0x4B000);
		hw_surface=fb1[0]+fb1_length;
		hwsurface_length=CAANOO_FB_LENGTH-fb1_length;
	}

	/* Set up the information for hardware surfaces */
	CAANOO_FreeHWSurfaces(this);
	CAANOO_InitHWSurfaces(this, current, hw_surface, hwsurface_length);

	current->w = 320;
	current->h = 240;
	current->pitch = SDL_CalculatePitch(current);
	if(bpp==8) current->flags|=SDL_HWPALETTE;

	/* Determine TV-out mode */
	this->tvout_mode=0;
	if(DPCCTRL0 & 0x8000) 
		this->tvout_mode |= MLC_PRI_ENABLED;
	if(DPCCTRL0_TV & 0x8000) 
		this->tvout_mode |= MLC_SEC_ENABLED;

	if(this->tvout_mode & MLC_PRI_ENABLED)
	{
		MLCLEFTRIGHT1_0=0;
		MLCTOPBOTTOM1_0=0;
		MLCLEFTRIGHT1_1=0;
		MLCTOPBOTTOM1_1=0;
		CAANOO_DirtyLayer1();
	
		/* Framebuffer layer settings */
		MLCBGCOLOR=0x000000;
		CAANOO_DirtyMLC();

		MLCLEFTRIGHT1=319;
		MLCTOPBOTTOM1=239;
		CAANOO_DirtyLayer1();

		MLCCONTROL1=mode;
		CAANOO_DirtyLayer1();

		if(bpp==8)
		{
			int i;
			for(i=0;i<256;i++)
				MLCPALETTE1=i<<24;
		}
		CAANOO_DirtyLayer1();

		MLCHSTRIDE1=bytesperpixel;
		MLCVSTRIDE1=current->pitch;
		CAANOO_DirtyLayer1();

		CAANOO_FlipFB(0);
		CAANOO_DirtyLayer1();
	}

	if(this->tvout_mode & MLC_SEC_ENABLED)
	{
		MLCLEFTRIGHT1_0_TV=0;
		MLCTOPBOTTOM1_0_TV=0;
		MLCLEFTRIGHT1_1_TV=0;
		MLCTOPBOTTOM1_1_TV=0;
		CAANOO_DirtyLayer1_TV();
	
		/* Framebuffer layer settings */
		MLCBGCOLOR_TV=0x000000;
		CAANOO_DirtyMLC_TV();

		MLCLEFTRIGHT1_TV=319;
		MLCTOPBOTTOM1_TV=239;
		CAANOO_DirtyLayer1_TV();

		MLCCONTROL1_TV=mode;
		CAANOO_DirtyLayer1_TV();

		if(bpp==8)
		{
			int i;
			for(i=0;i<256;i++)
				MLCPALETTE1_TV=i<<24;
		}
		CAANOO_DirtyLayer1_TV();

		MLCHSTRIDE1_TV=bytesperpixel;
		MLCVSTRIDE1_TV=current->pitch;
		CAANOO_DirtyLayer1_TV();

		CAANOO_FlipFB_TV(0);
		CAANOO_DirtyLayer1_TV();

		if(this->tvout_mode & MLC_PRI_ENABLED)
			this->FlipHWSurface = CAANOO_FlipHWSurface_Both;
		else
			this->FlipHWSurface = CAANOO_FlipHWSurface_TV;
	}

	usleep(100000);

	/* Set the update rectangle function */
	this->UpdateRects = CAANOO_DirectUpdate;

	/* We're done */
	return(current);
}

static int CAANOO_InitHWSurfaces(_THIS, SDL_Surface *screen, unsigned char *base, int size)
{
	vidmem_bucket *bucket;

	hwsurface_memleft = size;
	if ( hwsurface_memleft > 0 ) {
		bucket = (vidmem_bucket *)SDL_malloc(sizeof(*bucket));
		if ( bucket == NULL ) {
			SDL_OutOfMemory();
			return(-1);
		}
		bucket->prev = &surfaces;
		bucket->used = 0;
		bucket->base = base;
		bucket->size = size;
		bucket->next = NULL;
	} else {
		bucket = NULL;
	}

	surfaces.prev = NULL;
	surfaces.used = 1;
	surfaces.base = screen->pixels;
	surfaces.size = (unsigned int)((long)base - (long)surfaces.base);
	surfaces.next = bucket;
	screen->hwdata = (struct private_hwdata *)&surfaces;
	return(0);
}

static void CAANOO_FreeHWSurfaces(_THIS)
{
	vidmem_bucket *bucket, *freeable;

	bucket = surfaces.next;
	while ( bucket ) {
		freeable = bucket;
		bucket = bucket->next;
		SDL_free(freeable);
	}
	surfaces.next = NULL;
}

static int CAANOO_AllocHWSurface(_THIS, SDL_Surface *surface)
{
	vidmem_bucket *bucket;
	int size;
	int extra;

	size = surface->h * surface->pitch;
#ifdef CAANOO_DEBUG
	fprintf(stderr, "Allocating bucket of %d bytes\n", size);
#endif

	/* Quick check for available mem */
	if ( size > hwsurface_memleft ) {
		SDL_SetError("Not enough video memory");
		return(-1);
	}

	/* Search for an empty bucket big enough */
	for ( bucket=&surfaces; bucket; bucket=bucket->next ) {
		if ( ! bucket->used && (size <= bucket->size) ) {
			break;
		}
	}
	if ( bucket == NULL ) {
		SDL_SetError("Video memory too fragmented");
		return(-1);
	}

	/* Create a new bucket for left-over memory */
	extra = (bucket->size - size);
	if ( extra ) {
		vidmem_bucket *newbucket;

#ifdef CAANOO_DEBUG
	fprintf(stderr, "Adding new free bucket of %d bytes\n", extra);
#endif
		newbucket = (vidmem_bucket *)SDL_malloc(sizeof(*newbucket));
		if ( newbucket == NULL ) {
			SDL_OutOfMemory();
			return(-1);
		}
		newbucket->prev = bucket;
		newbucket->used = 0;
		newbucket->base = bucket->base+size;
		newbucket->size = extra;
		newbucket->next = bucket->next;
		if ( bucket->next ) {
			bucket->next->prev = newbucket;
		}
		bucket->next = newbucket;
	}

	/* Set the current bucket values and return it! */
	bucket->used = 1;
	bucket->size = size;
#ifdef CAANOO_DEBUG
	fprintf(stderr, "Allocated %d bytes at %p\n", bucket->size, bucket->base);
#endif
	hwsurface_memleft -= size;
	surface->flags |= SDL_HWSURFACE;
	surface->pixels = (void*)bucket->base;
	surface->hwdata = (struct private_hwdata *)bucket;
	memset(surface->pixels,0,size);
	return(0);
}

static void CAANOO_FreeHWSurface(_THIS, SDL_Surface *surface)
{
	vidmem_bucket *bucket, *freeable;

	/* Look for the bucket in the current list */
	for ( bucket=&surfaces; bucket; bucket=bucket->next ) {
		if ( bucket == (vidmem_bucket *)surface->hwdata ) {
			break;
		}
	}
	if ( bucket && bucket->used ) {
		/* Add the memory back to the total */
#ifdef CAANOO_DEBUG
	fprintf(stderr,"Freeing bucket of %d bytes\n", bucket->size);
#endif
		hwsurface_memleft += bucket->size;

		/* Can we merge the space with surrounding buckets? */
		bucket->used = 0;
		if ( bucket->next && ! bucket->next->used ) {
#ifdef CAANOO_DEBUG
	fprintf(stderr,"Merging with next bucket, for %d total bytes\n", bucket->size+bucket->next->size);
#endif
			freeable = bucket->next;
			bucket->size += bucket->next->size;
			bucket->next = bucket->next->next;
			if ( bucket->next ) {
				bucket->next->prev = bucket;
			}
			SDL_free(freeable);
		}
		if ( bucket->prev && ! bucket->prev->used ) {
#ifdef CAANOO_DEBUG
	fprintf(stderr,"Merging with previous bucket, for %d total bytes\n", bucket->prev->size+bucket->size);
#endif
			freeable = bucket;
			bucket->prev->size += bucket->size;
			bucket->prev->next = bucket->next;
			if ( bucket->next ) {
				bucket->next->prev = bucket->prev;
			}
			SDL_free(freeable);
		}
	}
	surface->pixels = NULL;
	surface->hwdata = NULL;
}

static int CAANOO_LockHWSurface(_THIS, SDL_Surface *surface)
{
	if ( surface == this->screen ) {
		SDL_mutexP(hw_lock);
	}
	return 0;
}
static void CAANOO_UnlockHWSurface(_THIS, SDL_Surface *surface)
{
	if ( surface == this->screen ) {
		SDL_mutexV(hw_lock);
	}
}

static int CAANOO_FlipHWSurface(_THIS, SDL_Surface *surface)
{
	if((surface->flags&SDL_VSYNCOFF)!=SDL_VSYNCOFF)
	{
		/* Wait for VSYNC */
		while(!(DPCCTRL0 & 0x400));

		/* Page flip */
		CAANOO_FlipFB(current_fb);
		CAANOO_DirtyLayer1();

		DPCCTRL0 |= 0x400; 
	}
	else
	{
		/* Page flip */
		CAANOO_FlipFB(current_fb);
		CAANOO_DirtyLayer1();
	}

	if(++current_fb>3) current_fb=0;
	surface->pixels=(void*)fb1[current_fb];
	return(0);
}

static int CAANOO_FlipHWSurface_TV(_THIS, SDL_Surface *surface)
{
	if((surface->flags&SDL_VSYNCOFF)!=SDL_VSYNCOFF)
	{
		/* Wait for VSYNC */
		while(!(DPCCTRL0_TV & 0x400));

		/* Page flip */
		CAANOO_FlipFB_TV(current_fb);
		CAANOO_DirtyLayer1_TV();

		DPCCTRL0_TV |= 0x400; 
	}
	else
	{
		/* Page flip */
		CAANOO_FlipFB_TV(current_fb);
		CAANOO_DirtyLayer1_TV();
	}

	if(++current_fb>3) current_fb=0;
	surface->pixels=(void*)fb1[current_fb];
	return(0);
}

static int CAANOO_FlipHWSurface_Both(_THIS, SDL_Surface *surface)
{
	if((surface->flags&SDL_VSYNCOFF)!=SDL_VSYNCOFF)
	{
		/* Wait for VSYNC */
		while(!(DPCCTRL0_TV & 0x400));

		/* Page flip */
		CAANOO_FlipFB(current_fb);
		CAANOO_DirtyLayer1();
		CAANOO_FlipFB_TV(current_fb);
		CAANOO_DirtyLayer1_TV();

		DPCCTRL0_TV |= 0x400; 
	}
	else
	{
		/* Page flip */
		CAANOO_FlipFB(current_fb);
		CAANOO_DirtyLayer1();
		CAANOO_FlipFB_TV(current_fb);
		CAANOO_DirtyLayer1_TV();
	}

	if(++current_fb>3) current_fb=0;
	surface->pixels=(void*)fb1[current_fb];
	return(0);
}

static void CAANOO_DirectUpdate(_THIS, int numrects, SDL_Rect *rects)
{
	/* The application is already updating the visible video memory */
	return;
}

/* Set hardware palette */
static int CAANOO_SetColors(_THIS, int firstcolor, int ncolors, SDL_Color *colors)
{
	int i;
	unsigned short color;

	if(this->tvout_mode & MLC_PRI_ENABLED)
	{
		for (i=firstcolor;i<ncolors;i++)
		{
			color=(colors[i].r>>3)<<11 | (colors[i].g>>2)<<5 | (colors[i].b>>3);
			MLCPALETTE1=i<<24 | color;
		}

		CAANOO_DirtyLayer1();
	}
	if(this->tvout_mode & MLC_SEC_ENABLED)
	{
		for (i=firstcolor;i<ncolors;i++)
		{
			color=(colors[i].r>>3)<<11 | (colors[i].g>>2)<<5 | (colors[i].b>>3);
			MLCPALETTE1_TV=i<<24 | color;
		}

		CAANOO_DirtyLayer1_TV();
	}

	return 0;
}


/* Note:  If we are terminated, this could be called in the middle of
   another SDL video routine -- notably UpdateRects.
*/
static void CAANOO_VideoQuit(_THIS)
{
	int i, j;

	if ( this->screen ) {
		/* Clear screen and tell SDL not to free the pixels */
// ghcstop fix to original: 041123
#if FB_INPUT_DEVICE_ENABLE
		if ( this->screen->pixels && CAANOO_InGraphicsMode(this) ) {
#else			
		if ( this->screen->pixels  ) {
#endif
			
#if defined(__powerpc__) || defined(__ia64__)	/* SIGBUS when using SDL_memset() ?? */
			Uint8 *rowp = (Uint8 *)this->screen->pixels;
			int left = this->screen->pitch*this->screen->h;
			while ( left-- ) { *rowp++ = 0; }
#else
			// GP2X WIZ jhkang 081006
			printf("FB_VideoQuit() h=%d, pitch=%d\n", this->screen->h, this->screen->pitch);

			/*SDL_memset(this->screen->pixels,0,this->screen->h*this->screen->pitch); */
#endif
		/* This test fails when using the VGA16 shadow memory */
			this->screen->pixels = NULL;
		}
	}

	/* Clear the lock mutex */
	if ( hw_lock ) {
		SDL_DestroyMutex(hw_lock);
		hw_lock = NULL;
	}

	/* Clean up defined video modes */
	if(SDL_modelist[0])
	{
		SDL_free(SDL_modelist[0]);
		SDL_modelist[0]=NULL;
	}

	/* Clean up the memory bucket list */
	CAANOO_FreeHWSurfaces(this);
	hw_surface=NULL;

	/* Clear screen */
	if(fb1[0])
	{
		memset((void*)fb1[0],0x00,0x4B000);
		if(fb1[1])
		{
			memset((void*)fb1[1],0x00,fb1_length);
			fb1[1]=NULL;
		}
		if(fb1[2])
		{
			memset((void*)fb1[2],0x00,fb1_length);
			fb1[2]=NULL;
		}
		if(fb1[3])
		{
			memset((void*)fb1[3],0x00,fb1_length);
			fb1[3]=NULL;
		}

		munmap((void*)fb1[0],CAANOO_FB_LENGTH);
		fb1[0]=NULL;
	}

	if(yuv_mem_start)
	{
		munmap((void*)yuv_mem_start,YUV_MEM_LENGTH);
		yuv_mem_start=NULL;
	}

	if(memregs32)
	{
		current_fb=0;
		if(this->tvout_mode & MLC_PRI_ENABLED)
		{
			/* Switch to system default framebuffer */
			CAANOO_FlipFB(0);
			CAANOO_DirtyLayer1();
		}
		if(this->tvout_mode & MLC_SEC_ENABLED)
		{
			/* Switch to system default framebuffer */
			CAANOO_FlipFB_TV(0);
			CAANOO_DirtyLayer1_TV();
		}	

		/* Restore backuped hw register values */
		if(bkregs32[0])
		{
			MLCADDRESS1=bkregs32[0][0];
			MLCCONTROL1=bkregs32[0][1];
			MLCLEFTRIGHT1=bkregs32[0][2];
			MLCTOPBOTTOM1=bkregs32[0][3];
			MLCLEFTRIGHT1_0=bkregs32[0][4];
			MLCTOPBOTTOM1_0=bkregs32[0][5];
			MLCLEFTRIGHT1_1=bkregs32[0][6];
			MLCTOPBOTTOM1_1=bkregs32[0][7];
			MLCBGCOLOR=bkregs32[0][8];
			MLCHSTRIDE1=bkregs32[0][9];
			MLCVSTRIDE1=bkregs32[0][10];
			MLCADDRESS2=bkregs32[0][11];
			MLCADDRESSCB=bkregs32[0][12];
			MLCADDRESSCR=bkregs32[0][13];
			MLCCONTROL2=bkregs32[0][14];
			MLCLEFTRIGHT2=bkregs32[0][15];
			MLCTOPBOTTOM2=bkregs32[0][16];
			MLCVSTRIDE2=bkregs32[0][17];
			MLCVSTRIDECB=bkregs32[0][18];
			MLCVSTRIDECR=bkregs32[0][19];
			MLCHSCALE=bkregs32[0][20];
			MLCVSCALE=bkregs32[0][21];

			CAANOO_DirtyLayer2();
			CAANOO_DirtyLayer1();
			CAANOO_DirtyMLC();

			free(bkregs32[0]);
			bkregs32[0]=NULL;
		}

		if(bkregs32[1])
		{
			MLCADDRESS1_TV=bkregs32[1][0];
			MLCCONTROL1_TV=bkregs32[1][1];
			MLCLEFTRIGHT1_TV=bkregs32[1][2];
			MLCTOPBOTTOM1_TV=bkregs32[1][3];
			MLCLEFTRIGHT1_0_TV=bkregs32[1][4];
			MLCTOPBOTTOM1_0_TV=bkregs32[1][5];
			MLCLEFTRIGHT1_1_TV=bkregs32[1][6];
			MLCTOPBOTTOM1_1_TV=bkregs32[1][7];
			MLCBGCOLOR_TV=bkregs32[1][8];
			MLCHSTRIDE1_TV=bkregs32[1][9];
			MLCVSTRIDE1_TV=bkregs32[1][10];
			MLCADDRESS2_TV=bkregs32[1][11];
			MLCADDRESSCB_TV=bkregs32[1][12];
			MLCADDRESSCR_TV=bkregs32[1][13];
			MLCCONTROL2_TV=bkregs32[1][14];
			MLCLEFTRIGHT2_TV=bkregs32[1][15];
			MLCTOPBOTTOM2_TV=bkregs32[1][16];
			MLCVSTRIDE2_TV=bkregs32[1][17];
			MLCVSTRIDECB_TV=bkregs32[1][18];
			MLCVSTRIDECR_TV=bkregs32[1][19];
			MLCHSCALE_TV=bkregs32[1][20];
			MLCVSCALE_TV=bkregs32[1][21];

			CAANOO_DirtyLayer2_TV();
			CAANOO_DirtyLayer1_TV();
			CAANOO_DirtyMLC_TV();

			free(bkregs32[1]);
			bkregs32[1]=NULL;
		}

		munmap((void*)memregs32,0x20000);
		memregs32=NULL;
	}
	memregs16=NULL;

	if(mem_fd>0) 
	{
		close(mem_fd);
		mem_fd=-1;
	}

	if(fb_fd>0) 
	{	
		close(fb_fd);
		fb_fd=-1;
	}

#if FB_INPUT_DEVICE_ENABLE	// ghcstop
	#if 0 // ghcstop_041123 delete, ̺κ ߿  
	FB_CloseMouse(this);
	#endif
//	dprintf("FB_closeKeyboard\n"); // ghcstop
	CAANOO_CloseKeyboard(this);
#endif	
	CAANOO_CloseMouse(this); //hyun
//printf("[4]FB_VideoQuit() h=%d, pitch=%d\n", this->screen->h, this->screen->pitch);
}

